"""Basic logger."""

import logging
import os


def get_logger(name: str) -> logging.Logger:
    """Get a logger.

    Parameters
    ----------
    name : str
        Name of the logger, typically __name__ is provided here

    Returns
    -------
    logging.Logger
        A logger
    """
    logger = logging.getLogger(name)
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        level=os.environ.get("LOGLEVEL", "INFO"),
    )
    return logger



"""Server framebuffer update message."""

from struct import unpack
from typing import List, Optional, Type

from wsvnc.encodings.encoding_interface import EncodingInterface
from wsvnc.encodings.raw_encoding import RawEncoding
from wsvnc.exceptions.encoding_exception import PixelEncodingError
from wsvnc.pixel_format import PixelFormat
from wsvnc.rectangle import Rectangle
from wsvnc.server_messages.server_message_interface import ServerMessage
from wsvnc.utils.logger import get_logger
from wsvnc.utils.safe_transport import SafeTransport

logger = get_logger(__name__)


class FrameBufferUpdate(ServerMessage):
    rectangles: List[Rectangle]
    encs: List[Type[EncodingInterface]]

    def __init__(
        self, pf: PixelFormat, encs: Optional[List[Type[EncodingInterface]]] = None
    ) -> None:
        self.pf = pf
        self.rectangles = []
        if encs:
            self.encs = encs
        else:
            self.encs = [RawEncoding]

    def type(self) -> int:
        return 0

    async def read(self, transport: SafeTransport, msg: bytes) -> None:
        """Handle a frame buffer update server message.

        Go through each rectangle and save the pixel data into an encoding object that
        will be later drawn on a PIL.Image object.
        Specified in RFC 6143 7.6.1

        Raise:
            PixelEncodingError: Couldn't decode rectangle.
        """
        # read off padding
        msg = msg[1:]

        # number of rectangles
        rect_num = unpack(">H", msg[0:2])[0]

        # read every rectangle
        msg = msg[2:]  # read off the rect number
        logger.debug(f"Num rectangles: {rect_num}")
        for _ in range(rect_num):
            rect = Rectangle()

            # wait for rectangle header if message is incomplete.
            msg = await transport.recvd(msg, 12)

            (rect.x, rect.y, rect.width, rect.height, enc_type) = unpack(
                "!HHHHi", msg[:12]
            )
            logger.debug(f"rectangle: {(rect.x, rect.y, rect.width, rect.height)}")
            msg = msg[12:]  # drop the rectangles header

            # here we read off the pixel data into the encodings colors field.
            success = False
            for enc in self.encs:
                encoding = enc()
                # use the first encoding that matches the enc_type 4 byte integer
                if encoding.type() == enc_type:
                    # update the msg first if necessary
                    msg = await encoding.fetch_additional_data(
                        rect.width, rect.height, transport, msg, self.pf
                    )
                    # remember to increase offset by data read
                    chg = encoding.read(rect.width, rect.height, msg, self.pf)
                    msg = msg[chg:]
                    rect.enc = encoding
                    success = True
            if not success:
                logger.error(
                    f"Rect: {rect.x, rect.y, rect.width, rect.height}, Encoding Type: {enc_type}"
                )
                raise PixelEncodingError(
                    "Server encoding does not match acceptable client encodings."
                )

            # next rectangle
            self.rectangles.append(rect)



"""VNC security class."""

from struct import unpack

from Cryptodome.Cipher import DES

from wsvnc.security.security_type_interface import SecurityTypeInterface
from wsvnc.utils.safe_transport import SafeTransport


class VNCSecurity(SecurityTypeInterface):
    def __init__(self, password: bytes):
        self.password = password

    def type(self) -> int:
        return 2

    async def handshake(self, transport: SafeTransport) -> None:
        """Implement basic VNC security protocol.

        Specified in RFC 6143 7.2.2
        """
        # read the challenge from the server (16 bytes)
        challenge_bytes = await transport.recv()
        if isinstance(challenge_bytes, bytes):
            challenge = unpack("!16s", challenge_bytes)[0]
        else:
            challenge = challenge_bytes

        # determine key
        password_length = len(self.password)
        key: bytes
        if password_length > 8:  # use only the first 8 bytes of the password
            key = self.password[:8]
        elif password_length == 8:
            key = self.password
        else:  # if key is smaller than 8 bytes we pad it to make it a 64-bit key.
            offset = 8 - password_length
            key += self.password + bytes([0x00] * offset)

        # reverse the bits
        key = bytes(
            sum((128 >> i) if (k & (1 << i)) else 0 for i in range(8)) for k in key
        )

        # now encrypt the challenge with our password and send to server
        des = DES.new(key=key, mode=DES.MODE_ECB)
        resp = des.encrypt(challenge)
        await transport.send(resp)



"""Handle color map entries message."""

from struct import unpack

from wsvnc.color import Color
from wsvnc.server_messages.server_message_interface import ServerMessage
from wsvnc.utils.safe_transport import SafeTransport


class ColorMapEntriesMessage(ServerMessage):
    _first_color: int
    _number_of_colors: int
    color_map: dict[int, Color]

    def __init__(self) -> None:
        self._first_color = 0
        self._number_of_colors = 0
        self.color_map = {}

    def type(self) -> int:
        return 1

    async def read(self, transport: SafeTransport, msg: bytes) -> None:
        """Handle Color Map Entries server message.

        Specified in RFC 6143 7.6.2
        """
        # read off passing.
        msg = msg[1:]

        # read off first color & num colors
        self._first_color = unpack(">H", msg[0:2])[0]
        self._number_of_colors = unpack(">H", msg[2:4])[0]

        # wait for the color map if message is incomplete.
        colors = msg[4:]
        expected_bytes = self._number_of_colors * 6
        colors = await transport.recvd(colors, expected_bytes)

        # read off each RGB color pixel value
        for i in range(self._number_of_colors):
            r, g, b = unpack(">HHH", colors[0:6])
            c = Color(r=r, g=g, b=b)
            self.color_map[i + self._first_color] = c
            colors = colors[6:]



"""Header encoding to enable JpegDiff Composition subencodings for TightPNG."""

from PIL import Image

from wsvnc.encodings.encoding_interface import EncodingInterface
from wsvnc.pixel_format import PixelFormat
from wsvnc.utils.safe_transport import SafeTransport


class TightPNGDiffCompEncoding(EncodingInterface):
    img: Image.Image

    async def fetch_additional_data(
        self,
        width: int,
        height: int,
        transport: SafeTransport,
        msg: bytes,
        pf: PixelFormat,
    ) -> bytes:
        return b""

    def type(self) -> int:
        return 1464686102

    def read(self, width: int, height: int, msg: bytes, pf: PixelFormat) -> int:
        """Header encoding."""
        return 0



"""Interface class for encodings."""

from abc import ABC, abstractmethod

from PIL.Image import Image

from wsvnc.pixel_format import PixelFormat
from wsvnc.utils.safe_transport import SafeTransport


class EncodingInterface(ABC):
    img: Image

    @abstractmethod
    def type(self) -> int:
        """Return the id of the encoding type.

        Returns
        -------
            int
        """
        pass

    @abstractmethod
    async def fetch_additional_data(
        self,
        width: int,
        height: int,
        transport: SafeTransport,
        msg: bytes,
        pf: PixelFormat,
    ) -> bytes:
        """Update rectangle data if necessary.

        Args:
            width (int): width of the rectangle
            height (int): height of the rectangle
            transport (SafeTransport): the socket
            msg (bytes): the data
            pf (PixelFormat): pixel format

        Returns
        -------
            bytes: The updated rectangle data.
        """
        pass

    @abstractmethod
    def read(self, width: int, height: int, msg: bytes, pf: PixelFormat) -> int:
        """Read the encoded pixels.

        Reads the pixels into self.img: Image, which is later assembled into
        the screen in the RFBClient.

        Args:
            width (int): width of the rectangle
            height (int): height of the rectangle
            msg (bytes): the encoded pixels in bytes
            pf (PixelFormat): the pre-set pixel format

        Returns
        -------
            int: number of encoded pixel bytes read
        """
        pass



"""Necessary to enable CopyRect encoding on ESXi."""

from PIL import Image

from wsvnc.encodings.encoding_interface import EncodingInterface
from wsvnc.pixel_format import PixelFormat
from wsvnc.utils.safe_transport import SafeTransport


class VMWDefineCursorEncoding(EncodingInterface):
    img: Image.Image

    async def fetch_additional_data(
        self,
        width: int,
        height: int,
        transport: SafeTransport,
        msg: bytes,
        pf: PixelFormat,
    ) -> bytes:
        # we need at least two bytes to determine the cursor type (second byte is garbage)
        cursor_header = await transport.recvd(msg, 2)

        # cursor type defined in the first byte
        self.cursor_type = cursor_header[0]

        # pixel length is fixed to the rectangle width * height * 4
        self.pixel_length = width * height * 4

        # mask length is either 0 or the pixel length
        if self.cursor_type == 0:
            self.mask_length = self.pixel_length
        else:
            self.mask_length = 0

        # now we must wait for pixel_length + mask_length bytes
        data = cursor_header[2:]
        data = await transport.recvd(data, self.pixel_length + self.mask_length)

        return data

    def type(self) -> int:
        return 1464686180

    def read(self, width: int, height: int, msg: bytes, pf: PixelFormat) -> int:
        """Not properly implemented.

        I believe this encoding gives you a base64 value from what I understand of the
        javascript in wmks, however this encoding is not actually used by the WebSDK.
        Instead, it just goes into the canvas style value and sets it to: `cursor:
        none`, to disable the canvas's cursor. I am going to leave this as unimplemented
        since it's not useful and only specific to ESXi, and I am only adding this
        encoding to enable CopyRect for ESXi.

        Also, apparently when this is received the FBU sends additional TightPNG
        rectangles with subencoding set to fill. I suppose that's meant to draw the
        cursor.
        """
        return self.pixel_length + self.mask_length



"""Standard No security type used by ESXi."""

from wsvnc.security.security_type_interface import SecurityTypeInterface
from wsvnc.utils.safe_transport import SafeTransport


class NoSecurity(SecurityTypeInterface):
    def type(self) -> int:
        return 1

    async def handshake(self, transport: SafeTransport) -> None:
        return



"""Basic raw pixel encoding class supported by ESXi."""

from struct import unpack

from PIL import Image

from wsvnc.encodings.encoding_interface import EncodingInterface
from wsvnc.pixel_format import PixelFormat
from wsvnc.utils.logger import get_logger
from wsvnc.utils.safe_transport import SafeTransport

logger = get_logger(__name__)


class RawEncoding(EncodingInterface):
    img: Image.Image

    def type(self) -> int:
        return 0

    async def fetch_additional_data(
        self,
        width: int,
        height: int,
        transport: SafeTransport,
        msg: bytes,
        pf: PixelFormat,
    ) -> bytes:
        """Fetch more pixel data if we don't have enough yet.

            This design is specific to raw encoding.
        Args:
            width: width of the rectangle
            height: height of the rectangle
            transport (SafeTransport): the socket
            msg (bytes): the existing (possibly incomplete) message.
            pf (PixelFormat): The pixel format

        Returns
        -------
            bytes: the updated message
        """
        return await transport.recvd(msg, (width * height * pf.bpp // 8))

    def read(self, width: int, height: int, msg: bytes, pf: PixelFormat) -> int:
        self.img = Image.new("RGBA", (width, height), (0, 0, 0, 0))

        bytes_per_pixel = pf.bpp // 8
        if pf.big_endian:
            byteorder = ">"
        else:
            byteorder = "<"

        offs = 0
        for i in range(height * width):
            x = i % width
            y = i // width

            pixel_bytes = msg[offs : offs + bytes_per_pixel]
            offs += bytes_per_pixel

            if len(pixel_bytes) < bytes_per_pixel:
                raise ValueError("Failed to read enough pixel bytes")

            if pf.bpp == 8:
                raw_pixel = pixel_bytes[0]
            elif pf.bpp == 16:
                raw_pixel = unpack(byteorder + "H", pixel_bytes)[0]
            elif pf.bpp == 32:
                raw_pixel = unpack(byteorder + "I", pixel_bytes)[0]

            # determines the color of this pixel
            # the final "&0xffff" ensures its a uint16 type
            if pf.true_color:
                r = (raw_pixel >> pf.red_shift & pf.red_max) & 0xFFFF
                g = (raw_pixel >> pf.green_shift & pf.green_max) & 0xFFFF
                b = (raw_pixel >> pf.blue_shift & pf.blue_max) & 0xFFFF
                self.img.putpixel((x, y), (r, g, b, 255))
            else:
                if pf.color_map is None:
                    logger.error("Color map is not ready. Cannot parse encoding.")
                    return offs
                color = pf.color_map[raw_pixel]
                self.img.putpixel((x, y), (color.r, color.g, color.b, 255))

        return offs



"""CopyRect encoding."""

from struct import unpack

from PIL import Image

from wsvnc.encodings.encoding_interface import EncodingInterface
from wsvnc.pixel_format import PixelFormat
from wsvnc.utils.safe_transport import SafeTransport


class CopyRectEncoding(EncodingInterface):
    img: Image.Image

    async def fetch_additional_data(
        self,
        width: int,
        height: int,
        transport: SafeTransport,
        msg: bytes,
        pf: PixelFormat,
    ) -> bytes:
        # CopyRect only contains 4 bytes (2 for src-x, and 2 for src-y)
        return await transport.recvd(msg, 4)

    def type(self) -> int:
        return 1

    def read(self, width: int, height: int, msg: bytes, pf: PixelFormat) -> int:
        """Determine the x & y of the original screen to copy.

        CopyRect is different than most encodings in that it requires the existing image
        to construct the new one. We can't do this at this function so here we just save
        the src-x & src-y value of the rectangle we're going to copy. In RFBClient
        _handleFBU() you'll see that the img value is assigned to a cropped section of
        the existing image based on these values.
        """
        (self.srcx, self.srcy) = unpack("!HH", msg[:4])
        return 4



"""Pixel format class representation."""

from struct import pack, unpack
from typing import Optional

from wsvnc.color import Color


class PixelFormat(object):
    bpp: int
    depth: int
    big_endian: int
    true_color: int
    red_max: int
    green_max: int
    blue_max: int
    red_shift: int
    green_shift: int
    blue_shift: int
    color_map: Optional[dict[int, Color]]

    def write_pixel_format(self) -> bytes:
        """Encode pixel format into bytes.

        Returns
        -------
            bytes: pixel format encoded into bytes
        """
        return pack(
            "!BBBBHHHBBBxxx",
            self.bpp,
            self.depth,
            self.big_endian,
            self.true_color,
            self.red_max,
            self.green_max,
            self.blue_max,
            self.red_shift,
            self.green_shift,
            self.blue_shift,
        )


def read_format(b: bytes) -> PixelFormat:
    """Read the pixel format.

    Args:
        b (bytes): 16 bytes that include the pixel format

    Returns
    -------
        PixelFormat: PixelFormat object
    """
    pf = PixelFormat()
    pf.bpp = b[0]
    pf.depth = b[1]
    pf.big_endian = b[2]
    pf.true_color = b[3]
    pf.red_max = unpack(">H", b[4:6])[0]
    pf.green_max = unpack(">H", b[6:8])[0]
    pf.blue_max = unpack(">H", b[8:10])[0]
    pf.red_shift = b[10]
    pf.green_shift = b[11]
    pf.blue_shift = b[12]
    return pf



"""Server message interface class."""

from abc import ABC, abstractmethod

from wsvnc.utils.safe_transport import SafeTransport


class ServerMessage(ABC):
    @abstractmethod
    def type(self) -> int:
        """Return message type (uint8).

        Returns
        -------
            int
        """
        pass

    @abstractmethod
    async def read(self, transport: SafeTransport, msg: bytes) -> None:
        """Handle bytes message.

        Args:
            transport (SafeTransport)
            msg (bytes)
        """
        pass



"""Wrapper to ensure binary transmission."""

from websockets import WebSocketClientProtocol


class SafeTransport:
    def __init__(self, transport: WebSocketClientProtocol) -> None:
        self.conn = transport

    async def recv(self) -> bytes:
        """Guarantee we receive bytes from connection."""
        data = await self.conn.recv()
        if isinstance(data, bytes):
            return data
        raise ValueError("Received data is not bytes.")

    async def recvd(self, msg: bytes, length: int) -> bytes:
        """Wait for a specific number of bytes given an existing message."""
        data = bytearray(msg)
        while len(data) < length:
            data.extend(await self.recv())
        return data

    async def send(self, msg: bytes) -> None:
        """Guarantee we send bytes.

        Args:
            msg (bytes): msg to be sent.
        """
        await self.conn.send(msg)



"""TightPNG encoding standard."""

from io import BytesIO

from PIL import Image

from wsvnc.encodings.encoding_interface import EncodingInterface
from wsvnc.pixel_format import PixelFormat
from wsvnc.utils.logger import get_logger
from wsvnc.utils.safe_transport import SafeTransport

logger = get_logger(__name__)


class TightPNGEncoding(EncodingInterface):
    img: Image.Image

    def __init__(self) -> None:
        self.sub_enc_jpeg = 144
        self.sub_enc_fill = 128
        self.sub_enc_png = 160
        self.sub_enc_diff_jpeg = 176
        self.sub_enc_mask = 240

    def type(self) -> int:
        return -260

    async def fetch_additional_data(
        self,
        width: int,
        height: int,
        transport: SafeTransport,
        msg: bytes,
        pf: PixelFormat,
    ) -> bytes:
        """Fetch more pixel data if we don't have enough yet.

            This design is specific to TightVNC Encoding.
        Args:
            width: width of the rectangle
            height: height of the rectangle
            transport (SafeTransport): the socket
            msg (bytes): the existing (possibly incomplete) message.
            pf (PixelFormat): The pixel format

        Returns
        -------
            bytes: the updated message
        """
        # if necessary wait for the TightPNG header (subencoding + length)
        pixel_data = await transport.recvd(msg, 4)

        # the subencoding is the first byte
        self.sub_encoding = pixel_data[0]
        logger.debug(f"Subencoding: {self.sub_encoding}")

        # unmask subencoding
        self.sub_encoding &= self.sub_enc_mask
        logger.debug(f"Subencoding Unmasked: {self.sub_encoding}")

        # if the sub_encoding is the color map we don't want to read off
        # any data aside from the subencoding
        if self.sub_encoding == self.sub_enc_fill:
            return pixel_data[1:]

        # get the length of the data encoded (lots of shifting here)
        # but this is correct (exact same logic in wmks.js from VMWare)
        data_length = pixel_data[1]
        data_length &= -129
        data_length += pixel_data[2] << 7
        data_length &= -16385
        data_length += pixel_data[3] << 14
        logger.debug(f"Tight Encoding length: {data_length}")

        self.data_length = data_length

        # chop off the header with subencoding & length
        pixel_data = pixel_data[4:]

        # now we know how many bytes we expect to receive, and so if we don't have them
        # we must wait
        pixel_data = await transport.recvd(pixel_data, data_length)

        return pixel_data

    def read(self, width: int, height: int, msg: bytes, pf: PixelFormat) -> int:
        # if sub_encoding is fill then our img is just a pixel
        # from three bytes in the buffer
        if self.sub_encoding == self.sub_enc_fill:
            self.img = Image.new("RGBA", (width, height), (msg[0], msg[1], msg[2], 255))
            return 3

        data = msg[: self.data_length]

        if (
            self.sub_encoding == self.sub_enc_jpeg
            or self.sub_encoding == self.sub_enc_png
        ):
            """Implements JPEG & PNG logic."""
            self.img = Image.open(BytesIO(data))

            if self.img.mode != "RGBA":
                self.img = self.img.convert("RGBA")

        if self.sub_encoding == self.sub_enc_diff_jpeg:
            """Implements JPEGDiff logic."""
            raise NotImplementedError

        return self.data_length



"""Server Bell message handler."""

from wsvnc.server_messages.server_message_interface import ServerMessage
from wsvnc.utils.safe_transport import SafeTransport


class BellMessage(ServerMessage):
    sig: int

    def type(self) -> int:
        return 2

    async def read(self, transport: SafeTransport, msg: bytes) -> None:
        self.sig = 0



"""Server cut text message."""

from struct import unpack

from wsvnc.server_messages.server_message_interface import ServerMessage
from wsvnc.utils.safe_transport import SafeTransport


class CutTextMessage(ServerMessage):
    cut_text: str

    def __init__(self) -> None:
        self.cut_text = ""

    def type(self) -> int:
        return 3

    async def read(self, transport: SafeTransport, msg: bytes) -> None:
        """Handle cut text server message.

        Specified in RFC 6143 7.6.4
        """
        # read off padding
        msg = msg[3:]

        # read text length
        text_len = unpack(">I", msg[0:4])[0]
        text = msg[4:]

        # wait for more text if message is incomplete.
        text = await transport.recvd(text, text_len)

        # read text
        self.cut_text = unpack("{l}s".format(l=text_len), text)[0]



"""Key constants for user to press special keys."""

from typing import List

supported_versions: List[bytes] = [b"RFB 003.008\n"]

KEY_BackSpace = 0xFF08
KEY_Tab = 0xFF09
KEY_Return = 0xFF0D
KEY_Escape = 0xFF1B
KEY_Insert = 0xFF63
KEY_Delete = 0xFFFF
KEY_Home = 0xFF50
KEY_End = 0xFF57
KEY_PageUp = 0xFF55
KEY_PageDown = 0xFF56
KEY_Left = 0xFF51
KEY_Up = 0xFF52
KEY_Right = 0xFF53
KEY_Down = 0xFF54
KEY_F1 = 0xFFBE
KEY_F2 = 0xFFBF
KEY_F3 = 0xFFC0
KEY_F4 = 0xFFC1
KEY_F5 = 0xFFC2
KEY_F6 = 0xFFC3
KEY_F7 = 0xFFC4
KEY_F8 = 0xFFC5
KEY_F9 = 0xFFC6
KEY_F10 = 0xFFC7
KEY_F11 = 0xFFC8
KEY_F12 = 0xFFC9
KEY_F13 = 0xFFCA
KEY_F14 = 0xFFCB
KEY_F15 = 0xFFCC
KEY_F16 = 0xFFCD
KEY_F17 = 0xFFCE
KEY_F18 = 0xFFCF
KEY_F19 = 0xFFD0
KEY_F20 = 0xFFD1
KEY_ShiftLeft = 0xFFE1
KEY_ShiftRight = 0xFFE2
KEY_ControlLeft = 0xFFE3
KEY_ControlRight = 0xFFE4
KEY_MetaLeft = 0xFFE7
KEY_MetaRight = 0xFFE8
KEY_AltLeft = 0xFFE9
KEY_AltRight = 0xFFEA

KEY_Scroll_Lock = 0xFF14
KEY_Sys_Req = 0xFF15
KEY_Num_Lock = 0xFF7F
KEY_Caps_Lock = 0xFFE5
KEY_Pause = 0xFF13
KEY_Super_L = 0xFFEB  # windows-key, apple command key
KEY_Super_R = 0xFFEC  # windows-key, apple command key
KEY_Hyper_L = 0xFFED
KEY_Hyper_R = 0xFFEE

KEY_KP_0 = 0xFFB0
KEY_KP_1 = 0xFFB1
KEY_KP_2 = 0xFFB2
KEY_KP_3 = 0xFFB3
KEY_KP_4 = 0xFFB4
KEY_KP_5 = 0xFFB5
KEY_KP_6 = 0xFFB6
KEY_KP_7 = 0xFFB7
KEY_KP_8 = 0xFFB8
KEY_KP_9 = 0xFFB9
KEY_KP_Enter = 0xFF8D

KEY_ForwardSlash = 0x002F
KEY_BackSlash = 0x005C
KEY_SpaceBar = 0x0020



"""Class to represent framebuffer update rectangles sent back by VNC server."""

from wsvnc.encodings.encoding_interface import EncodingInterface


class Rectangle(object):
    x: int  # 2 bytes
    y: int
    width: int
    height: int
    enc: EncodingInterface



"""VNC Client class, the main entrypoint for agents."""

from __future__ import annotations

import asyncio
import threading
import time
from io import BytesIO
from ssl import SSLContext
from types import TracebackType
from typing import List, Optional, Type

import websockets
from PIL import Image

from wsvnc.encodings.encoding_interface import EncodingInterface
from wsvnc.pixel_format import PixelFormat
from wsvnc.rfb.rfb_client import RFBClient
from wsvnc.security import no_security, security_type_interface
from wsvnc.server_messages.bell import BellMessage
from wsvnc.utils.logger import get_logger

logger = get_logger(__name__)


class WSVNCClient:
    def __init__(
        self,
        ticket_url: str,
        ssl_context: Optional[SSLContext] = None,
        origin: str = "http://localhost",
        security_type: security_type_interface.SecurityTypeInterface = no_security.NoSecurity(),
        keep_screen_updated: bool = False,
        shared_flag: int = 1,
    ) -> None:
        self.ticket_url = ticket_url
        self.origin = origin
        self.security_type = security_type
        self.ssl_context = ssl_context
        self.shared_flag = shared_flag
        # event loop
        self._loop = asyncio.new_event_loop()
        # start the client
        self._handshake_done = threading.Event()
        self._caught_exception: Exception | None = None
        self._run()
        self._handshake_done.wait(30)

        # if we have an exception during initial
        # connection we will throw it here
        # this will notify the user and prevent them
        # from accidentally using the client when it never connected
        if self._caught_exception:
            raise self._caught_exception

        if keep_screen_updated:
            self.set_resend_flag()

    def set_resend_flag(self, on: bool = True) -> None:
        """Set FBUR resend flag.

        If this is set to True, the screen will continually update in the background.

        Args:
            on (bool): Resend flag value.
        """
        self._rfb_client.resend_flag = on
        if on:
            self.update_screen()

    def set_pixel_format(self, pf: PixelFormat) -> None:
        """Set the pixel formatting the server will use.

        Args:
            pf (PixelFormat): Pixel format object
        """
        asyncio.run_coroutine_threadsafe(
            self._rfb_client.set_pixel_format(pf), self._loop
        )

    def set_encodings(self, encs: List[Type[EncodingInterface]]) -> None:
        """Set the encodings the client will use.

        Args:
            encs (List[Type[EncodingInterface]]): encodings that implement the interface.
        """
        asyncio.run_coroutine_threadsafe(
            self._rfb_client.set_encodings(encs), self._loop
        )

    def send_key(self, key: int) -> None:
        """Press a key then release.

        Args:
            key (int): Key ID
        """
        self.key_event(key, True)
        time.sleep(0.1)
        self.key_event(key, False)

    def move(self, xpos: int, ypos: int) -> None:
        """Move mouse to position(xpos,ypos).

        Used before any other click action to ensure cursor is in the correct position.

        Args:
            xpos (int)
            ypos (int)
        """
        self.pointer_event(xpos, ypos, 0)

    def release(self, xpos: int, ypos: int) -> None:
        """Relase is the same as move().

        Args:
            xpos (int)
            ypos (int)
        """
        self.move(xpos, ypos)

    def left_click(self, xpos: int, ypos: int) -> None:
        """Left click at position(xpos, ypos).

        Args:
            xpos (int)
            ypos (int)
        """
        self.move(xpos, ypos)
        self.pointer_event(xpos, ypos, 1)
        self.pointer_event(xpos, ypos, 0)

    def double_left_click(self, xpos: int, ypos: int) -> None:
        """Double left click at position(xpos, ypos).

        Args:
            xpos (int)
            ypos (int)
        """
        self.move(xpos, ypos)
        self.left_click(xpos, ypos)
        time.sleep(0.05)
        self.left_click(xpos, ypos)

    def press(self, xpos: int, ypos: int) -> None:
        """Left-Press and then hold.

        Args:
            xpos (int)
            ypos (int)
        """
        self.move(xpos, ypos)
        self.pointer_event(xpos, ypos, 1)

    def right_click(self, xpos: int, ypos: int) -> None:
        """Right click at position(xpos, ypos).

        Args:
            xpos (int)
            ypos (int)
        """
        self.move(xpos, ypos)
        self.pointer_event(xpos, ypos, 4)
        self.pointer_event(xpos, ypos, 0)

    def wheel_up(self, xpos: int, ypos: int, delay_ms: int = 50) -> None:
        """Scrolls up at position(xpos, ypos) for duration delay_ms.

        Args:
            xpos (int)
            ypos (int)
            delay_ms (int, optional): time to scroll up. Defaults to 50.
        """
        self.pointer_event(xpos, ypos, 8)
        time.sleep(delay_ms * 0.001)
        self.pointer_event(xpos, ypos, 0)

    def wheel_down(self, xpos: int, ypos: int, delay_ms: int = 50) -> None:
        """Scrolls down at position(xpos, ypos) for duration delay_ms.

        Args:
            xpos (int)
            ypos (int)
            delay_ms (int, optional): time to scroll down. Defaults to 50.
        """
        self.pointer_event(xpos, ypos, 16)
        time.sleep(delay_ms * 0.001)
        self.pointer_event(xpos, ypos, 0)

    def wheel(self, xpos: int, ypos: int, delay_ms: int, down: bool = False) -> None:
        """Use mouse wheel.

        If down is set to true the mouse wheel will scroll down, if not it will scroll up.

        Args:
            xpos (int)
            ypos (int)
            delay_ms (int): time mouse wheel will scroll in milliseconds
            down (bool): true=wheel down, false=wheel up
        """
        if down:
            self.wheel_down(xpos, ypos, delay_ms)
        else:
            self.wheel_up(xpos, ypos, delay_ms)

    def click_and_drag(self, xpos: int, ypos: int, newx: int, newy: int) -> None:
        """Clicks at position(xpos, ypos) and hold to position(newx, newy).

        Args:
            xpos (int): starting x coord.
            ypos (int): starting y coord.
            newx (int): end x coord.
            newy (int): end y coord.
        """
        self.move(xpos, ypos)
        self.pointer_event(xpos, ypos, 1)
        self.pointer_event(newx, newy, 1)

    def emit_text(self, text: str) -> None:
        """Type text on the server. These should be strictly [a-Z0-9].

        Args:
            text (str): text to be emitted.
        """
        for chr in text:
            if self._is_shift_required(chr):
                self.key_event(65505, True)  # press shift
                time.sleep(0.1)
                self.send_key(ord(chr))
                self.key_event(65505, False)  # release shift
            else:
                self.send_key(ord(chr))

    def cut_text(self, text: str) -> None:
        """Client tells the server it has text in its clipboard.

        Args:
            text (str): _description_
        """
        asyncio.run_coroutine_threadsafe(self._rfb_client.cut_text(text), self._loop)

    def _is_shift_required(self, c: str) -> bool:
        """Determine if we should press shift when typing this letter.

        Does not check for caps lock, so don't set caps lock.

        Args:
            c (str): a character

        Returns
        -------
            bool: true if it shift is needed on the keyboard, false otherwise
        """
        if c.isupper():
            return True

        special = '!@#$%^&*()_+{}|:"<>?'
        if c in special:
            return True

        return False

    def update_screen(
        self,
        width: Optional[int] = None,
        height: Optional[int] = None,
        incremental: bool = False,
        x: int = 0,
        y: int = 0,
    ) -> None:
        """Send a remote framebuffer update request that will update the screen.

        If resend flag is False, this will run once in the background.
        If no parameters are set then the FBUR will request the entire screen.

        Args:
            width (Optional[int]): Width of the frame. Defaults to None.
            height (Optional[int]): Height of the frame. Defaults to None.
            incremental (bool, optional): Incremental flag (see RFC 7.5.3). Defaults to False.
            x (int, optional): Starting x coord of frame. Defaults to 0.
            y (int, optional): Starting y coord of frame. Defaults to 0.
        """
        if width is None:
            width = self._rfb_client.width
        if height is None:
            height = self._rfb_client.height
        asyncio.run_coroutine_threadsafe(
            self._rfb_client.framebuffer_update_request(
                x, y, width, height, incremental
            ),
            self._loop,
        )

    def key_event(self, key: int, down: bool) -> None:
        """Non-abstracted key event call.

        Use this command if you need something beyond the key event abstractions.

        Args:
            key (int): Key ID
            down (bool): press down
        """
        asyncio.run_coroutine_threadsafe(
            self._rfb_client.key_event(key, down), self._loop
        )

    def pointer_event(self, xpos: int, ypos: int, mask: int) -> None:
        """Non-abstracted pointer event call.

        Use this command if you need something beyond the pointer event abstractions.

        Args:
            xpos (int)
            ypos (int)
            mask (int): the mouse button ID
        """
        asyncio.run_coroutine_threadsafe(
            self._rfb_client.pointer_event(xpos, ypos, mask), self._loop
        )

    def get_screen(self) -> Image.Image | None:
        """Return latest update of the screen.

        Will return None if screen was never updated with update_screen().

        Returns
        -------
            Image: The RGBA image of the screen
        """
        return self._rfb_client.img

    def get_screen_bytes(self) -> bytes | None:
        """Return latest update of the screen in bytes.

        Will return None if screen was never updated with update_screen().

        Returns
        -------
            bytes: Image of the screen in byte format
        """
        buffer = BytesIO()
        img = self.get_screen()
        if img is not None:
            img.save(buffer, format="PNG")
            return buffer.getvalue()

        return None

    def get_clipboard(self) -> str:
        """Return the clipboard of the client.

        Returns
        -------
            str: clipboard of the client
        """
        return self._rfb_client.clipboard

    def get_pixel_format(self) -> PixelFormat:
        """Return the current PixelFormat in use.

        Returns
        -------
            PixelFormat: PixelFormat in use
        """
        return self._rfb_client.pixel_format

    def get_server_name(self) -> str:
        """Return the name of the server.

        Returns
        -------
            str: Server name
        """
        return self._rfb_client.server_name

    def get_encodings(self) -> List[Type[EncodingInterface]]:
        """Return the encodings we are using.

        Returns
        -------
            List[Type[EncodingInterface]]: List of encodings in use
        """
        return self._rfb_client.encs

    def get_bell(self) -> BellMessage | None:
        """Return a bell message if we received one.

        Returns
        -------
            BellMessage | None: A bell message if received by the client
        """
        return self._rfb_client.bell

    async def _main_loop(self) -> None:
        """Loop will keep the websocket connection alive and listen for messages."""
        try:
            async with websockets.connect(
                self.ticket_url,
                ssl=self.ssl_context,
                origin=self.origin,  # type: ignore
                subprotocols=["binary"],  # type: ignore
                ping_interval=None,
                max_size=2**25,
                read_limit=2**25,
            ) as conn:
                self._rfb_client = RFBClient(conn, self.security_type, self.shared_flag)
                await self._rfb_client.handshake()
                self._handshake_done.set()
                await self._rfb_client.listen()  # handles exceptions internally
                logger.info("VNC Client no longer listening")
        except Exception as e:
            # this block runs if the websocket connection fails for any reason
            # will set the handshake so the thread doesn't block
            # can only trigger during __init__()
            self._caught_exception = e
            self._handshake_done.set()
            raise

    async def _close(self) -> None:
        """Close the websocket connection."""
        await self._rfb_client.close()

    def _run(self) -> None:
        """Start the thread to run an asynchronous connection over a websocket."""
        logger.info("Starting VNC Client")
        asyncio.set_event_loop(self._loop)

        # create the connection task
        self._loop.create_task(self._main_loop())

        # thread is a variable now.
        self._thread = threading.Thread(target=self._loop.run_forever, daemon=True)
        self._thread.start()

    def close(self) -> None:
        """Terminates the thread and closes the websocket connection."""
        # takes a moment to shut down everything, wait one second after closing websocket connection.
        asyncio.run_coroutine_threadsafe(self._close(), self._loop)
        time.sleep(1)
        self._loop.call_soon_threadsafe(self._loop.stop)

        self._thread.join()
        logger.info("Client shutdown complete.")

    def __enter__(self) -> WSVNCClient:
        """Context manager enter method.

        Returns
        -------
            WSVNCClient: The client
        """
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> bool:
        """Context manager exit method.

        Args:
            exc_type (Optional[Type[BaseException]]): exception type
            exc_val (Optional[BaseException]): exception value
            exc_tb (Optional[TracebackType]): exception traceback

        Returns
        -------
            bool: client closed (in)correctly
        """
        if exc_type is not None:
            logger.error(
                f"Exception type: {exc_type}\n Exception value: {exc_val}\n Exception traceback: {exc_tb}"
            )
            return False
        self.close()
        return True



"""Security interface class."""

from abc import ABC, abstractmethod

from wsvnc.utils.safe_transport import SafeTransport


class SecurityTypeInterface(ABC):
    @abstractmethod
    def type(self) -> int:
        """Return security type (uint8).

        Returns
        -------
            int
        """
        pass

    @abstractmethod
    async def handshake(self, transport: SafeTransport) -> None:
        """Handle security handshake.

        Args:
            transport (SafeTransport): the websocket connection.
        """
        pass



"""Client that implements RFB Protocol according to RFC 6143."""

import re
import traceback
from struct import pack, unpack
from typing import List, Optional, Type

from PIL import Image
from websockets import WebSocketClientProtocol

from wsvnc.constants import supported_versions
from wsvnc.encodings.copyrect_encoding import CopyRectEncoding
from wsvnc.encodings.encoding_interface import EncodingInterface
from wsvnc.encodings.pseudo_desktop_size_encoding import PseudoDesktopSizeEncoding
from wsvnc.encodings.raw_encoding import RawEncoding
from wsvnc.encodings.vmware_define_cursor_encoding import VMWDefineCursorEncoding
from wsvnc.pixel_format import PixelFormat, read_format
from wsvnc.security.no_security import NoSecurity
from wsvnc.security.security_type_interface import SecurityTypeInterface
from wsvnc.server_messages.bell import BellMessage
from wsvnc.server_messages.color_map_entries import ColorMapEntriesMessage
from wsvnc.server_messages.cut_text import CutTextMessage
from wsvnc.server_messages.framebuffer_update import FrameBufferUpdate
from wsvnc.utils.logger import get_logger
from wsvnc.utils.safe_transport import SafeTransport

logger = get_logger(__name__)


class RFBClient:
    encs: List[Type[EncodingInterface]]
    width: int
    height: int
    pixel_format: PixelFormat
    _server_name_length: int
    server_name: str
    img: Optional[Image.Image] = None
    resend_flag: bool = False
    bell: Optional[BellMessage] = None

    def __init__(
        self,
        conn: WebSocketClientProtocol,
        security_type: SecurityTypeInterface = NoSecurity(),
        shared_flag: int = 1,
    ) -> None:
        self.transport = SafeTransport(conn)
        self.security_type = security_type
        self.encs = [RawEncoding]
        self.clipboard = ""
        self.shared_flag = shared_flag

    """ Functions that are sent by the VNC Client -> ESXi """

    async def cut_text(self, text: str) -> None:
        """Tell the server that the client has new text in its cut buffer.

        RFC 6143 Section 7.5.6.

        Args:
            text (str): string to copy
        """
        self.clipboard = text
        await self.transport.send(pack("!BxxxI", 6, len(text)) + text.encode("utf-8"))

    async def framebuffer_update_request(
        self, x: int, y: int, width: int, height: int, incremental: bool = False
    ) -> None:
        """Request a framebuffer update from the server.

        There may be an indefinite time
        between the request and the actual framebuffer update being received.
        RFC 6143 Section 7.5.3.

        Args:
            x (int): x-pos start of frame
            y (int): y-pos start of frame
            width (int): width of frame
            height (int): height of frame
            incremental (bool): true = incremental,
        """
        logger.debug("sent frame buffer update request.")
        await self.transport.send(pack("!BBHHHH", 3, incremental, x, y, width, height))

    async def key_event(self, key: int, down: bool) -> None:
        """Press or release key on server.

        Many keys are there ASCII equivalent, but some are unique and in constants.
        RFC 6143 Section 7.5.4.

        Args:
            key (int): keyId #
            down (bool): true=down, false=up.
        """
        if down:
            send_down = 1
        else:
            send_down = 0
        await self.transport.send(pack("!BBxxI", 4, send_down, key))

    async def pointer_event(self, x: int, y: int, mask: int) -> None:
        """Send pointer event to server.

        RFC 6143 Section 7.5.5.

        Args:
            x (int): x-cord on screen
            y (int): y-cord on screen
            mask (int): button to press (0 to unpress) (1=leftclick) (4=rightclick)
        """
        await self.transport.send(pack("!BBHH", 5, mask, x, y))

    async def set_encodings(self, encs: List[Type[EncodingInterface]]) -> None:
        """Set encoding types that can be sent from the server.

        Args:
            encs (List[EncodingInterface]): List of implemented encodings
        """
        msg = pack("!BxH", 2, len(encs))
        for enc in encs:
            encoding = enc()
            msg += pack(">i", encoding.type())
        await self.transport.send(msg)
        self.encs = encs + [RawEncoding]

    async def set_pixel_format(self, format: PixelFormat) -> None:
        """Set pixel format for FramebufferUpdate messages from the server.

        RFC 6143 Section 7.5.1.

        Args:
            format (PixelFormat): _description_
        """
        self.pixel_format = format
        logger.debug("Setting Pixel Format.")
        await self.transport.send(pack("!Bxxx", 0) + format.write_pixel_format()[1:])

    async def close(self) -> None:
        """Close the websocket connection."""
        await self.transport.conn.close()

    """ Functions that handle messages received by the Client, Server -> Client """

    async def listen(self) -> None:
        """Async function that listens and handles server messages."""
        try:
            async for message in self.transport.conn:
                if not isinstance(message, bytes):
                    raise TypeError("Client should receive bytes.")
                if len(message) == 0:
                    logger.error("Received a message from the server with no bytes.")
                else:
                    logger.debug(f"Received: {len(message)} bytes")
                    if (
                        message[0] == 0
                    ):  # if first byte is 0 we received a framebuffer_update
                        logger.debug("received framebuffer update.")
                        message = await self.transport.recvd(
                            message, 16
                        )  # read the rest of the FBU.
                        await self._handle_framebuffer_update(message)
                        logger.debug("updated screen!")
                        if self.resend_flag:
                            await self.framebuffer_update_request(
                                0, 0, self.width, self.height, True
                            )
                    elif message[0] == 1:
                        logger.debug("received colorMap message.")
                        message = await self.transport.recvd(
                            message, 6
                        )  # read the rest of the FBU. # read rest of color map
                        await self._handle_color_map(message)
                    elif message[0] == 2:
                        logger.debug("received bell message.")
                        await self._handle_bell()
                    elif message[0] == 3:
                        logger.debug("received server cut text message.")
                        message = await self.transport.recvd(
                            message, 8
                        )  # read the rest of the FBU. # read rest of cut text
                        await self._handle_server_cut_text(message)
        except Exception as e:
            logger.error("RFBClient Encountered An exception! {s}".format(s=e))
            logger.error(f"RFBClient Exception Traceback: {traceback.format_exc()}")
            await self.close()

    async def _handle_framebuffer_update(self, msg: bytes) -> None:
        """Async function helper to handle framebuffer update messages from server."""
        if self.img is None:  # set the image to the screen size and all black.
            self.img = Image.new("RGBA", (self.width, self.height), (0, 0, 0, 0))

        fbu = FrameBufferUpdate(self.pixel_format, self.encs)
        await fbu.read(self.transport, msg[1:])
        logger.debug("Updating image with pixels.")
        for rect in fbu.rectangles:
            enc = rect.enc
            logger.debug(f"Processing: {rect.width * rect.height}, pixels")

            # VMWare define unimplemented
            if isinstance(enc, VMWDefineCursorEncoding):
                continue

            # DesktopSize is special since it defines the screen size
            # If we get this encoding we should redefine the img to be blank
            # and to whatever size it now should be, and then send an FBUR for
            # the entire screen to get an update ASAP
            if isinstance(enc, PseudoDesktopSizeEncoding):
                logger.info(
                    f"Redefining Desktop Size to: width: {rect.width}, height: {rect.height}"
                )
                self.img = Image.new("RGBA", (rect.width, rect.height), (0, 0, 0, 0))
                self.width = rect.width
                self.height = rect.height
                await self.framebuffer_update_request(
                    0, 0, rect.width, rect.height, False
                )
                return

            # CopyRect is special since we need the existing image to crop from
            # to assemble the new one
            if isinstance(enc, CopyRectEncoding):
                enc.img = self.img.crop(
                    (enc.srcx, enc.srcy, enc.srcx + rect.width, enc.srcy + rect.height)
                )

            self.img.paste(enc.img, (rect.x, rect.y))

    async def _handle_server_cut_text(self, msg: bytes) -> None:
        """Handle server cut text messages.

        RFC 6143 section 7.6.4. Replaces the rfb_client.cut_text field.
        """
        sct = CutTextMessage()
        await sct.read(self.transport, msg[1:])
        logger.debug(f"Client has cut text in buffer: {sct.cut_text}")
        self.clipboard = sct.cut_text

    async def _handle_color_map(self, msg: bytes) -> None:
        """Handle color map entries server message.

        RFC 6143 section 7.6.2. Sets the pixel_format.color_map field that will be used
        when decoding pixels.
        """
        cme = ColorMapEntriesMessage()
        await cme.read(self.transport, msg[1:])
        self.pixel_format.color_map = cme.color_map

    async def _handle_bell(self) -> None:
        """Handle server bell message.

        If you want the bell to do something I recommend overriding the read() in
        BellMessage.
        """
        self.bell = BellMessage()
        await self.bell.read(self.transport, b"")

    """ Helpers to establish the VNC connection over websocket."""

    async def handshake(self) -> None:
        """Do initial handshake with VNC server."""
        msg = await self.transport.recv()
        if re.search(rb"RFB (\d+\.\d+)\n", msg):
            logger.debug(f"Received initial handshake message: {msg!r}")
        else:
            logger.error("Received bad handshake message!")
            raise (ValueError("VNC handshake start failed!"))

        if self._validate_security(msg):
            # we'll support 003.008 (the latest security)
            await self.transport.send(b"RFB 003.008\n")

            # now we read the servers security types
            # first byte is the length, the rest are the types supported
            # for ESXi, only None is supported (1).
            security_types_recv = await self.transport.recv()
            security_types = security_types_recv[1:]
            # now we do the security handshake
            await self._security_handshake(security_types)

            # security handshake result (4 bytes should be 0, else is error)
            security_result_recv = await self.transport.recv()
            security_result = unpack(">I", security_result_recv)[0]
            if security_result != 0:
                raise (ValueError("Security handshake failed!"))

            # send share flag
            await self.transport.send(pack("!B", self.shared_flag))

            # ServerInit
            server_init_recv = await self.transport.recv()
            # the first 4 bytes of the initialization are the width and height of the screen respectively
            self.width = unpack(">H", server_init_recv[0:2])[0]
            self.height = unpack(">H", server_init_recv[2:4])[0]
            # the next 16 bytes are the pixel format
            self.pixel_format = read_format(server_init_recv[4:20])
            # the next 4 bytes is the LENGTH of the NAME of the server
            self._server_name_length = unpack(">I", server_init_recv[20:24])[0]
            # finally read in the server name.
            self.server_name = str(
                unpack(
                    "{l}s".format(l=self._server_name_length),
                    server_init_recv[-self._server_name_length :],
                )[0]
            )

            logger.info("Handshake done!")
            logger.info("Client Ready!")
        else:
            logger.error("Received unsupported security version.")

    async def _security_handshake(self, sec_types: bytes) -> None:
        """Commence security handshake.

        First determines if the server handles the
        security type we want to use. Then, commences the security handshake.

        Args:
            sec_types (bytes): list of security types the server can handle (ESXi handles only 1 (noSecurity)).
        """
        for sec_type in sec_types:
            if self.security_type.type() == sec_type:
                await self.transport.send(pack("!B", sec_type))
                # print("sent security type.")
                await self.security_type.handshake(self.transport)
            else:
                raise (UserWarning("Server does not support desired security type!"))

    def _validate_security(self, text: bytes) -> bool:
        """Validate the server is using security we support.

        Args:
            text (bytes): initial server message

        Returns
        -------
            bool: true if we support, false if we don't (kills handshake)
        """
        if text in supported_versions:
            return True
        else:
            return False
